#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#include <assert.h>
#include "dhnn.h"

#if defined(unix) || defined(__unix__)
void usleep(unsigned long usec);
#else
void Sleep(unsigned long msec);
#endif

int slow_processing = 0;

struct vec {
	float *v;
	int size;

	int _dyn_alloc;
};

static struct vec *create_vec(int sz);
static void init_vec(struct vec *v, int sz);
static void free_vec(struct vec *v);
static float dot(struct vec a, struct vec b);

struct sqmat {
	float **m;
	int size;
	
	int _dyn_alloc;
};

static struct sqmat *create_sqmat(int sz);
static void init_sqmat(struct sqmat *m, int sz);
static void free_sqmat(struct sqmat *m);
static void get_column(struct vec *v, struct sqmat m, int idx);
static void mul(struct sqmat dmat, struct vec a, struct vec b);
static void add(struct sqmat dmat, struct sqmat smat);
static void zero_diagonal(struct sqmat mat);

struct dhnn {
	int size;
	struct sqmat weights;
};

void *dh_create(int nodes) {
	struct dhnn *dh;
	
	if(!(dh = malloc(sizeof(struct dhnn)))) {
		return 0;
	}

	init_sqmat(&dh->weights, nodes);
	dh->size = nodes;

	return dh;
}

void dh_free(void *dh) {
	free_sqmat(&((struct dhnn*)dh)->weights);
	free(dh);
}

int dh_get_node_count(void *dh) {
	return ((struct dhnn*)dh)->size;
}

float dh_get_memory(void *dh) {
	int sz = ((struct dhnn*)dh)->size;
	return sz / (2.0 * log(sz));
}

void dh_train(void *dhnn, int *pattern) {
	int i;
	struct sqmat tmp_mat;
	struct vec vpat;
	struct dhnn *dh = dhnn;

	init_sqmat(&tmp_mat, dh->size);
	init_vec(&vpat, dh->size);

	for(i=0; i<dh->size; i++) {
		vpat.v[i] = (float)pattern[i];
	}

	mul(tmp_mat, vpat, vpat);
	zero_diagonal(tmp_mat);
	add(dh->weights, tmp_mat);

	free_vec(&vpat);
	free_sqmat(&tmp_mat);
}

static int converged(int *ndif, int sz) {
	int i;
	for(i=0; i<sz; i++) {
		if(ndif[i] != 0) {
			return 0;
		}
	}
	return 1;
}

static float calc_energy(void *dhnn, struct vec v) {
	int i, j;
	struct dhnn *dh = dhnn;
	
	float sum = 0.0;
	for(j=0; j<dh->size; j++) {
		for(i=0; i<dh->size; i++) {
			sum += v.v[j] * v.v[i] * dh->weights.m[i][j];
		}
	}
	return sum * -0.5;
}

void dh_process(void *dhnn, int *out, int *pattern) {
	int i;
	struct vec vpat, tmp_vec;
	struct dhnn *dh = dhnn;
	int *ndif = malloc(dh->size * sizeof(int));
	
	memset(ndif, INT_MAX, dh->size * sizeof(int));

	init_vec(&vpat, dh->size);
	init_vec(&tmp_vec, dh->size);
	
	/* feed the input to the neurons (as the virtual previous output) */
	for(i=0; i<dh->size; i++) {
		vpat.v[i] = (float)pattern[i];
	}

	do {
		float new_out;
		int nidx = rand() % dh->size;
			
		get_column(&tmp_vec, dh->weights, nidx);
		new_out = dot(vpat, tmp_vec) >= 0.0 ? 1 : -1;

		ndif[nidx] = vpat.v[nidx] - new_out;
		vpat.v[nidx] = new_out;

		printf("Energy: %.2f     \r", calc_energy(dhnn, vpat));
		fflush(stdout);

		/* slow down processing to be able to watch the changing energy */
		if(slow_processing) {
#if defined(unix) || defined(__unix__)
			usleep(100000);
#else
			Sleep(100);
#endif
		}
	} while(!converged(ndif, dh->size));
	putchar('\n');

	/* copy the result to the output array */
	for(i=0; i<dh->size; i++) {
		out[i] = (int)vpat.v[i];
	}
}

void dh_set_slow_processing(int enable) {
	slow_processing = enable;
}

/* -------- helper math functions ------- */

static struct vec *create_vec(int sz) {
	struct vec *v = malloc(sizeof(struct vec));
	if(!v) return 0;

	v->v = malloc(sz * sizeof(float));
	memset(v->v, 0, sz * sizeof(float));
	v->size = sz;
	v->_dyn_alloc = 1;

	return v;
}

static void init_vec(struct vec *v, int sz) {
	v->v = malloc(sz * sizeof(float));
	memset(v->v, 0, sz * sizeof(float));
	v->size = sz;
	v->_dyn_alloc = 0;
}

static void free_vec(struct vec *v) {
	free(v->v);
	if(v->_dyn_alloc) free(v);
}

static float dot(struct vec a, struct vec b) {
	int i;
	float res = 0.0;

	assert(a.size == b.size);
	
	for(i=0; i<a.size; i++) {
		res += a.v[i] * b.v[i];
	}
	return res;
}



static struct sqmat *create_sqmat(int sz) {
	struct sqmat *m = malloc(sizeof(struct sqmat));
	if(!m) return 0;
	
	init_sqmat(m, sz);
	m->_dyn_alloc = 1;
	return m;
}


static void init_sqmat(struct sqmat *m, int sz) {
	int i;
	m->m = malloc(sz * sizeof(float*));
	for(i=0; i<sz; i++) {
		m->m[i] = malloc(sz * sizeof(float));
		memset(m->m[i], 0, sz * sizeof(float));
	}
	m->size = sz;
	m->_dyn_alloc = 0;
}

static void free_sqmat(struct sqmat *m) {
	int i;
	for(i=0; i<m->size; i++) {
		free(m->m[i]);
	}
	free(m->m);
	if(m->_dyn_alloc) free(m);
}

static void get_column(struct vec *v, struct sqmat m, int idx) {
	int i;
	for(i=0; i<v->size; i++) {
		v->v[i] = m.m[i][idx];
	}
}

static void mul(struct sqmat dmat, struct vec a, struct vec b) {
	int i, j, sz;

	assert(dmat.size == a.size && dmat.size == b.size);
	sz = dmat.size;

	for(i=0; i<sz; i++) {
		for(j=0; j<sz; j++) {
			dmat.m[i][j] = a.v[i] * b.v[j];
		}
	}
}

static void add(struct sqmat dmat, struct sqmat smat) {
	int i, j, sz;
	
	assert(dmat.size == smat.size);
	sz = dmat.size;

	for(i=0; i<sz; i++) {
		for(j=0; j<sz; j++) {
			dmat.m[i][j] += smat.m[i][j];
		}
	}
}

static void zero_diagonal(struct sqmat mat) {
	int i;
	for(i=0; i<mat.size; i++) {
		mat.m[i][i] = 0.0;
	}
}
