#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include "dhnn.h"

struct pattern {
	int *p;
	int sz;
};

int defer_size(const char *fname);
int training(const char *fname);
int *get_input(int sz);
struct pattern parse_pattern(const char *pstr);


#define BUFSIZE	2048
char line[BUFSIZE];

void *nnet;

const char *usage =
	"dhnn accepts a pattern from stdin and tries to associate it with a known\n"
	"pattern from the training set (provided in a training file).\n\n"
	"Usage: ./dhnn [options]\n"
	"options:\n"
	"\t-s: set the network size (default: defer from the training set)\n"
	"\t-t: set the training data file (default: \"training_set\")\n"
	"\t-d: delay recall processing to be able to see the energy changing\n"
	"\t-h: show this help text\n\n";

int main(int argc, char **argv) {
	int i, size = -1;
	int *res;
	char *tset_fname = "training_set";
	int *input;

	/* parse command line options ... */
	for(i=1; i<argc; i++) {
		if(argv[i][0] == '-' && argv[i][2] == 0) {
			switch(argv[i][1]) {
			case 's':
				if(!isdigit(argv[++i][0])) {
					fprintf(stderr, "-s must be followed by a number (size)\n");
					return EXIT_FAILURE;
				}
				size = atoi(argv[i]);
				break;

			case 't':
				tset_fname = argv[++i];
				break;

			case 'd':
				dh_set_slow_processing(1);
				break;

			default:
				fprintf(stderr, "unrecognized argument \"%s\"\n", argv[i]);
				return EXIT_FAILURE;
			}
		} else {
			fprintf(stderr, "unrecognized argument \"%s\"\n", argv[i]);
			return EXIT_FAILURE;
		}
	}

	/* create the nnet */
	if(size <= 0) {
		fprintf(stderr, "trying to defer the nnet size from the training set...\n");
		if((size = defer_size(tset_fname)) <= 0) {
			if(!size) {
				fprintf(stderr, "invalid network size... (0)\n");
			}
			return EXIT_FAILURE;
		}
		fprintf(stderr, "nnet size: %d\n", size);
	}
	
	if(!(nnet = dh_create(size))) {
		return EXIT_FAILURE;
	}
	fprintf(stderr, "calculated neural network memory: %f\n", dh_get_memory(nnet));

	/* training */
	if(training(tset_fname) == -1) {
		return EXIT_FAILURE;
	}

	/* try and associate the input to one of the training set patterns */
	res = malloc(size * sizeof(int));
	if(!(input = get_input(size))) {
		return EXIT_FAILURE;
	}
	
	printf("input: ");
	for(i=0; i<size; i++) {
		printf("%d ", input[i]);
	}
	putchar('\n');
	
	dh_process(nnet, res, input);
	free(input);
	
	printf("output: ");
	for(i=0; i<size; i++) {
		printf("%d ", res[i]);
	}
	putchar('\n');
	free(res);

	/* destroy the net and exit */
	dh_free(nnet);
	return 0;
}

int defer_size(const char *fname) {
	FILE *fp;
	struct pattern p;
	
	if(!(fp = fopen(fname, "r"))) {
		fprintf(stderr, "failed to open the training set file \"%s\": %s\n", fname, strerror(errno));
		return -1;
	}

	if(!fgets(line, BUFSIZE, fp)) {
		fprintf(stderr, "the training set file is empty\n");
		fclose(fp);
		return -1;
	}
	p = parse_pattern(line);

	free(p.p);
	fclose(fp);
	return p.sz;
}

int training(const char *fname) {
	FILE *fp;
	int pcount = 0;

	if(!(fp = fopen(fname, "r"))) {
		fprintf(stderr, "failed to open the training set file \"%s\": %s\n", fname, strerror(errno));
		return -1;
	}

	while(fgets(line, BUFSIZE, fp)) {
		struct pattern p;
		fprintf(stderr, "training: %s", line);
		
		p = parse_pattern(line);
		if(dh_get_node_count(nnet) != p.sz) {
			fprintf(stderr, "inconsistent training set!!!\n");
			fclose(fp);
			return -1;
		}

		dh_train(nnet, p.p);	/* train this pattern */
		free(p.p);
		pcount++;
	}

	fclose(fp);
	return pcount;
}

int *get_input(int sz) {
	struct pattern p;
	if(!fgets(line, BUFSIZE, stdin)) {
		fprintf(stderr, "EOF while reading input pattern\n");
	}

	p = parse_pattern(line);
	if(p.sz != sz) {
		free(p.p);
		fprintf(stderr, "inconsistency between training set and input sizes\n");
		return 0;
	}
	
	return p.p;
}

/* ----- pattern parsing ----- */

struct pelem {
	int elem;
	struct pelem *next;
};

struct pattern parse_pattern(const char *pstr) {
	struct pattern p;
	struct pelem *plist, *tmp;
	char *tok = 0, *strptr;
	int *pptr;

	strptr = malloc(strlen(pstr) + 1);
	strcpy(strptr, pstr);
	
	if(!(plist = malloc(sizeof(struct pelem)))) {
		perror("failed to allocate temporary pattern element list");
		exit(EXIT_FAILURE);
	}
	plist->next = 0;
	
	p.sz = 0;
	while((tok = strtok(tok ? 0 : strptr, ", \t\n"))) {
		if(isdigit(tok[0]) || (tok[0] == '-' && isdigit(tok[1]))) {
			struct pelem *node = malloc(sizeof(struct pelem));
			if(!node) {
				perror("memory allocation failure");
				exit(EXIT_FAILURE);
			}
			node->elem = atoi(tok);
			node->next = plist->next;
			plist->next = node;
		}
		p.sz++;
	}

	/* got all the elements, now allocate the actual pattern, and copy the
	 * elements in reverse order, and destroy the list.
	 */
	if(!(p.p = malloc(p.sz * sizeof(int)))) {
		perror("memory allocation failure");
		exit(EXIT_FAILURE);
	}

	/* get rid of the dummy node */
	tmp = plist;
	plist = plist->next;
	free(tmp);

	/* copy the elements, while erasing the list */
	pptr = p.p + p.sz - 1;
	while(plist) {
		*pptr-- = plist->elem;
		tmp = plist;
		plist = plist->next;
		free(tmp);
	}

	return p;
}
